@@ -19,6 +19,12 @@
 //
 //
 require_once('session.php');
+require_once('User.php');
+require_once('Group_Permission.php');
+require_once('Monitor_Permission.php');
+require_once('User_Role.php');
+require_once('Role_Group_Permission.php');
+require_once('Role_Monitor_Permission.php');
 require_once(__DIR__.'/../vendor/autoload.php');
 use \Firebase\JWT\JWT;
 
@@ -39,15 +45,15 @@ function password_type($password) {
 
 // this function migrates mysql hashing to bcrypt, if you are using PHP >= 5.5
 // will be called after successful login, only if mysql hashing is detected
-function migrateHash($username, $password) {
+function migrateHash($user, $pass) {
   if ( function_exists('password_hash') ) {
-    global $user;
-    ZM\Info("Migrating $username to bcrypt scheme");
+    ZM\Info("Migrating $user to bcrypt scheme");
     // let it generate its own salt, and ensure bcrypt as PASSWORD_DEFAULT may change later
     // we can modify this later to support argon2 etc as switch to its own password signature detection 
-    $bcrypt_hash = password_hash($password, PASSWORD_BCRYPT);
-    dbQuery('UPDATE Users SET Password=? WHERE Username=?', array($bcrypt_hash, $username));
-    $user['Password'] = $bcrypt_hash;
+    $bcrypt_hash = password_hash($pass, PASSWORD_BCRYPT);
+    //ZM\Info ("hased bcrypt $pass is $bcrypt_hash");
+    $update_password_sql = 'UPDATE Users SET Password=\''.$bcrypt_hash.'\' WHERE Username=\''.$user.'\'';
+    dbQuery($update_password_sql);
     # Since password field has changed, existing auth_hash is no longer valid
     generateAuthHash(ZM_AUTH_HASH_IPS, true);
   } else {
@@ -59,9 +65,9 @@ function migrateHash($username, $password) {
 // core function used to load a User record by username and password
 function validateUser($username='', $password='') {
   if (ZM_CASE_INSENSITIVE_USERNAMES) {
-      $sql = 'SELECT * FROM Users WHERE Enabled=1 AND LOWER(Username)=LOWER(?)';
+    $sql = 'SELECT * FROM Users WHERE Enabled=1 AND LOWER(Username)=LOWER(?)';
   } else {
-      $sql = 'SELECT * FROM Users WHERE Enabled=1 AND Username=?';
+    $sql = 'SELECT * FROM Users WHERE Enabled=1 AND Username=?';
   }
   // local user, shouldn't affect the global user
   $user = dbFetchOne($sql, NULL, array($username)); // Not global
@@ -100,14 +106,14 @@ function validateUser($username='', $password='') {
   } // switch password_type
 
   if ($password_correct) {
-    return array($user, 'OK');
+    return array(new ZM\User($user), 'OK');
   }
   return array(false, "Login denied for user \"$username\"");
 } # end function validateUser
 
 function userLogout() {
   global $user;
-  ZM\Info('User "'.($user?$user['Username']:'no one').'" logged out');
+  ZM\Info('User "'.($user?$user->Username():'no one').'" logged out');
   $user = null;// unset only clears the local variable
   zm_session_clear();
 }
@@ -131,6 +137,8 @@ function validateToken($token, $allowed_token_type='access') {
       ZM\Error("Token type mismatch. Expected $allowed_token_type but got $type");
       return array(false, 'Incorrect token type');
     }
+  } else {
+    ZM\Debug('Not comparing token types as [any] was passed');
   }
   
   $username = $jwt_payload['user'];
@@ -150,7 +158,7 @@ function validateToken($token, $allowed_token_type='access') {
       $user = null;// unset only clears the local variable
       return array(false, 'Token revoked. Please re-generate');
     }
-    $user = $saved_user_details;
+    $user = new ZM\User($saved_user_details);
     return array($user, 'OK');
   }
   ZM\Error("Could not retrieve user $username details");
@@ -186,7 +194,7 @@ function getAuthUser($auth) {
         $authHash = md5($authKey);
 
         if ($auth == $authHash) {
-          return $user;
+          return new ZM\User($user);
         } // end if $auth == $authHash
       } // end foreach hour
     } // end foreach user
@@ -207,72 +215,279 @@ function getAuthUser($auth) {
           $authHash = md5($authKey);
 
           if ($auth == $authHash) {
-            return $user;
+            return new ZM\User($user);
           } // end if $auth == $authHash
         } // end foreach hour
       } // end foreach user
     } // end if 
   } // end if using auth hash
 
-  ZM\Error("Unable to authenticate user from auth hash '$auth'");
+  ZM\Info("Unable to authenticate user from auth hash '$auth'");
   return null;
 } // end getAuthUser($auth)
 
-function calculateAuthHash($remoteAddr) {
+function calculateAuthHash($remoteAddr='') {
   global $user;
   $local_time = localtime();
-  $authKey = ZM_AUTH_HASH_SECRET.$user['Username'].$user['Password'].$remoteAddr.$local_time[2].$local_time[3].$local_time[4].$local_time[5];
+  $authKey = ZM_AUTH_HASH_SECRET.$user->Username().$user->Password().$remoteAddr.$local_time[2].$local_time[3].$local_time[4].$local_time[5];
   #ZM\Debug("Generated using hour:".$local_time[2] . ' mday:' . $local_time[3] . ' month:'.$local_time[4] . ' year: ' . $local_time[5] );
   return md5($authKey);
 }
 
 function generateAuthHash($useRemoteAddr, $force=false) {
   global $user;
-  if (ZM_OPT_USE_AUTH and (ZM_AUTH_RELAY == 'hashed') and isset($user['Username']) and isset($user['Password']) and isset($_SESSION)) {
-    $time = time();
-
-    # We use 1800 so that we regenerate the hash at half the TTL
-    $mintime = $time - (ZM_AUTH_HASH_TTL * 1800);
-
-    # Appending the remoteAddr prevents us from using an auth hash generated for a different ip
-    if ($force or ( !isset($_SESSION['AuthHash'.$_SESSION['remoteAddr']]) ) or ( $_SESSION['AuthHashGeneratedAt'] < $mintime )) {
-      $auth = calculateAuthHash($useRemoteAddr?$_SESSION['remoteAddr']:'');
-      # Don't both regenerating Auth Hash if an hour hasn't gone by yet
-      $_SESSION['AuthHash'.$_SESSION['remoteAddr']] = $auth;
-      $_SESSION['AuthHashGeneratedAt'] = $time;
-      # Because we don't write out the session, it shouldn't actually get written out to disk.  However if it does, the GeneratedAt should protect us.
-    } # end if AuthHash is not cached
-    return $_SESSION['AuthHash'.$_SESSION['remoteAddr']];
+  if (!isset($_SESSION['remoteAddr'])) $_SESSION['remoteAddr'] = '';
+  if (ZM_OPT_USE_AUTH and (ZM_AUTH_RELAY == 'hashed') and $user and $user->Username()) {
+    if (!isset($_SESSION)) {
+      # Appending the remoteAddr prevents us from using an auth hash generated for a different ip
+      #$auth = calculateAuthHash($useRemoteAddr?$_SESSION['remoteAddr']:'');
+      $auth = calculateAuthHash();
+      return $auth;
+    } else {
+      $time = time();
+      # We use 1800 so that we regenerate the hash at half the TTL
+      $mintime = $time - (ZM_AUTH_HASH_TTL * 1800);
+      $remoteAddr = ZM_AUTH_HASH_IPS ? $_SESSION['remoteAddr'] : '';
+      # Appending the remoteAddr prevents us from using an auth hash generated for a different ip
+      if ($force or ( !isset($_SESSION['AuthHash'.$remoteAddr]) ) or ( $_SESSION['AuthHashGeneratedAt'] < $mintime )) {
+        $auth = calculateAuthHash($useRemoteAddr ? $remoteAddr : '');
+        # Don't both regenerating Auth Hash if an hour hasn't gone by yet
+        $_SESSION['AuthHash'.$remoteAddr] = $auth;
+        $_SESSION['AuthHashGeneratedAt'] = $time;
+        # Because we don't write out the session, it shouldn't actually get written out to disk.  However if it does, the GeneratedAt should protect us.
+      } # end if AuthHash is not cached
+      return $_SESSION['AuthHash'.$remoteAddr];
+    }
   } # end if using AUTH and AUTH_RELAY
   return '';
 }
 
+$group_permissions = null;
+$monitor_permissions = null; # hash indexed by MonitorId
+$role_group_permissions = null;
+$role_monitor_permissions = null;
+
 function visibleMonitor($mid) {
   global $user;
+  if (!$user) return false;
 
-  return ( $user && empty($user['MonitorIds']) || in_array($mid, explode(',', $user['MonitorIds'])) );
+  global $monitor_permissions;
+
+  # First check for direct monitor permission
+  if ($monitor_permissions === null) {
+    $monitor_permissions = array_to_hash_by_key('MonitorId', $user->Monitor_Permissions());
+  }
+
+  if (isset($monitor_permissions[$mid])) {
+    $permission = $monitor_permissions[$mid]->Permission();
+    if ($permission != 'Inherit') {
+      ZM\Debug('Returning '.($permission == 'None' ? false : true)." for monitor $mid from user monitor permission");
+      return ($permission == 'None' ? false : true);
+    }
+  }
+
+  global $group_permissions;
+  if ($group_permissions === null)
+    $group_permissions = $user->Group_Permissions();
+
+  # Check user's group permissions
+  $group_permission_value = 'Inherit';
+  foreach ($group_permissions as $permission) {
+    $value = $permission->MonitorPermission($mid);
+    if ($value == 'None') {
+      ZM\Debug("Can't view monitor $mid because of user group ".$permission->Group()->Name().' '.$permission->Permission());
+      return false;
+    } else if ($value == 'View' or $value == 'Edit') {
+      $group_permission_value = $value;
+    }
+  }
+  if ($group_permission_value != 'Inherit') return true;
+
+  # Check role permissions if user has a role
+  $role = $user->Role();
+  if ($role) {
+    global $role_monitor_permissions;
+    if ($role_monitor_permissions === null) {
+      $role_monitor_permissions = array_to_hash_by_key('MonitorId', $role->Monitor_Permissions());
+    }
+
+    if (isset($role_monitor_permissions[$mid])) {
+      $permission = $role_monitor_permissions[$mid]->Permission();
+      if ($permission != 'Inherit') {
+        ZM\Debug('Returning '.($permission == 'None' ? false : true)." for monitor $mid from role monitor permission");
+        return ($permission == 'None' ? false : true);
+      }
+    }
+
+    global $role_group_permissions;
+    if ($role_group_permissions === null)
+      $role_group_permissions = $role->Group_Permissions();
+
+    # Check role's group permissions
+    $role_group_permission_value = 'Inherit';
+    foreach ($role_group_permissions as $permission) {
+      $value = $permission->MonitorPermission($mid);
+      if ($value == 'None') {
+        ZM\Debug("Can't view monitor $mid because of role group ".$permission->Group()->Name().' '.$permission->Permission());
+        return false;
+      } else if ($value == 'View' or $value == 'Edit') {
+        $role_group_permission_value = $value;
+      }
+    }
+    if ($role_group_permission_value != 'Inherit') return true;
+
+    # Check role's base Monitors permission if user's is 'None'
+    if ($user->Monitors() == 'None' and $role->Monitors() != 'None') {
+      ZM\Debug("Using role Monitors permission ".$role->Monitors()." for monitor $mid");
+      return true;
+    }
+  }
+
+  return ($user->Monitors() != 'None');
 }
 
 function canView($area, $mid=false) {
   global $user;
+  if (!$user) return false;
+  if ($mid) return visibleMonitor($mid);
 
-  return ( $user && ($user[$area] == 'View' || $user[$area] == 'Edit') && ( !$mid || visibleMonitor($mid) ) );
+  # Check user's direct permission first
+  if ($user->$area() && ($user->$area() != 'None')) {
+    return true;
+  }
+
+  # If user's permission is 'None', check role permission
+  $role = $user->Role();
+  if ($role && $role->$area() && ($role->$area() != 'None')) {
+    return true;
+  }
+
+  return false;
+}
+
+function editableMonitor($mid) {
+  global $user;
+  if (!$user) {
+    ZM\Debug("Not logged in");
+    return false;
+  }
+
+  global $monitor_permissions;
+
+  # First check for direct monitor permission
+  if ($monitor_permissions === null) {
+    $monitor_permissions = array_to_hash_by_key('MonitorId', ZM\Monitor_Permission::find(array('UserId'=>$user->Id())));
+  }
+  if (isset($monitor_permissions[$mid])) {
+    $permission = $monitor_permissions[$mid]->Permission();
+    if ($permission == 'None' or $permission == 'View') {
+      return false;
+    } else if ($permission == 'Edit') {
+      return true;
+    }
+  }
+
+  global $group_permissions;
+  if ($group_permissions === null)
+    $group_permissions = ZM\Group_Permission::find(array('UserId'=>$user->Id()));
+
+  # Check user's group permissions
+  foreach ($group_permissions as $permission) {
+    $perm_value = $permission->MonitorPermission($mid);
+    if ($perm_value == 'Edit') {
+      return true;
+    }
+  }
+
+  # Check role permissions if user has a role
+  $role = $user->Role();
+  if ($role) {
+    global $role_monitor_permissions;
+    if ($role_monitor_permissions === null) {
+      $role_monitor_permissions = array_to_hash_by_key('MonitorId', $role->Monitor_Permissions());
+    }
+
+    if (isset($role_monitor_permissions[$mid])) {
+      $permission = $role_monitor_permissions[$mid]->Permission();
+      if ($permission == 'None' or $permission == 'View') {
+        return false;
+      } else if ($permission == 'Edit') {
+        return true;
+      }
+    }
+
+    global $role_group_permissions;
+    if ($role_group_permissions === null)
+      $role_group_permissions = $role->Group_Permissions();
+
+    # Check role's group permissions
+    foreach ($role_group_permissions as $permission) {
+      $perm_value = $permission->MonitorPermission($mid);
+      if ($perm_value == 'Edit') {
+        return true;
+      }
+    }
+
+    # Check role's base Monitors permission if user's doesn't allow edit
+    $user_monitors = $user->Monitors();
+    if ($user_monitors != 'Edit' && $user_monitors != 'Create') {
+      $role_monitors = $role->Monitors();
+      if ($role_monitors == 'Edit' || $role_monitors == 'Create') {
+        return true;
+      }
+    }
+  }
+
+  return (($user->Monitors() == 'Edit') || ($user->Monitors() == 'Create'));
 }
 
 function canEdit($area, $mid=false) {
   global $user;
 
-  return ( $user && ($user[$area] == 'Edit') && ( !$mid || visibleMonitor($mid) ));
+  if (!$user) return false;
+  if ($mid) return editableMonitor($mid);
+
+  # Check user's direct permission first
+  if ($user->$area() == 'Edit' or $user->$area() == 'Create') {
+    return true;
+  }
+
+  # If user's permission doesn't allow edit, check role permission
+  $role = $user->Role();
+  if ($role && ($role->$area() == 'Edit' or $role->$area() == 'Create')) {
+    return true;
+  }
+
+  return false;
+}
+
+function canCreate($area) {
+  global $user;
+  if (!$user) return false;
+
+  # Check user's direct permission first
+  if ($user->$area() == 'Create') {
+    return true;
+  }
+
+  # If user's permission doesn't allow create, check role permission
+  $role = $user->Role();
+  if ($role && $role->$area() == 'Create') {
+    return true;
+  }
+
+  return false;
 }
 
 function userFromSession() {
   $user = null; // Not global
   if (isset($_SESSION['username'])) {
+    $remoteAddr = ZM_AUTH_HASH_IPS ? $_SESSION['remoteAddr'] : '';
     if (ZM_AUTH_HASH_LOGINS and (ZM_AUTH_RELAY == 'hashed')) {
       # Extra validation, if logged in, then the auth hash will be set in the session, so we can validate it.
       # This prevent session modification to switch users
-      if (isset($_SESSION['AuthHash'.$_SESSION['remoteAddr']]))
-        $user = getAuthUser($_SESSION['AuthHash'.$_SESSION['remoteAddr']]);
+      if (isset($_SESSION['AuthHash'.$remoteAddr]))
+        $user = getAuthUser($_SESSION['AuthHash'.$remoteAddr]);
       else
         ZM\Debug('No auth hash in session, there should have been');
     } else {
@@ -282,7 +497,7 @@ function userFromSession() {
       } else {
         $sql = 'SELECT * FROM Users WHERE Enabled=1 AND Username=?';
       }
-      $user = dbFetchOne($sql, NULL, array($_SESSION['username']));
+      $user = new ZM\User(dbFetchOne($sql, NULL, array($_SESSION['username'])));
     }
   }
   return $user;
@@ -294,9 +509,13 @@ function get_auth_relay() {
       return 'auth='.generateAuthHash(ZM_AUTH_HASH_IPS);
     } else if (ZM_AUTH_RELAY == 'plain') {
       // password probably needs to be escaped
-      return 'username='.$_SESSION['username'].'&password='.urlencode($_SESSION['password']);
+      return 'username='.(isset($_SESSION['username'])?$_SESSION['username']:'').'&password='.urlencode(isset($_SESSION['password']) ? $_SESSION['password'] : '');
     } else if (ZM_AUTH_RELAY == 'none') {
-      return 'username='.$_SESSION['username'];
+      if (isset($_SESSION['username'])) {
+        return 'username='.$_SESSION['username'];
+      } else {
+        ZM\Warning("ZM_AUTH_RELAY==none but there is no username in session.");
+      }
     } else {
       ZM\Error('Unknown value for ZM_AUTH_RELAY ' . ZM_AUTH_RELAY);
     }
@@ -316,10 +535,28 @@ if (ZM_OPT_USE_AUTH) {
 
     if (ZM_AUTH_HASH_LOGINS && empty($user) && !empty($_REQUEST['auth'])) {
       $user = getAuthUser($_REQUEST['auth']);
+      if ($user) {
+        $remoteAddr = ZM_AUTH_HASH_IPS ? $_SESSION['remoteAddr'] : '';
+        if (isset($_SESSION['AuthHash'.$remoteAddr]) and ($_SESSION['AuthHash'.$remoteAddr] != $_REQUEST['auth'])) {
+          unset($_SESSION['AuthHashGeneratedAt']);
+          unset($_SESSION['AuthHash'.$remoteAddr]);
+        }
+        $_SESSION['username'] = $user->Username();
+      }
+    } else if (!(empty($_REQUEST['user']) or empty($_REQUEST['pass']))) {
+      # The shortened versions are used in auth_relay = PLAIN
+      $ret = validateUser($_REQUEST['user'], $_REQUEST['pass']);
+      if (!$ret[0]) {
+        ZM\Warning($ret[1]);
+        unset($user); // unset should be ok here because we aren't in a function
+        return;
+      }
+      $user = $ret[0];
     } else if (!(empty($_REQUEST['username']) or empty($_REQUEST['password']))) {
+      # Longer versions are used on login page
       $ret = validateUser($_REQUEST['username'], $_REQUEST['password']);
       if (!$ret[0]) {
-        ZM\Error($ret[1]);
+        ZM\Warning($ret[1]);
         unset($user); // unset should be ok here because we aren't in a function
         return;
       }
@@ -370,7 +607,7 @@ if (ZM_OPT_USE_AUTH) {
       $password = $_REQUEST['password'];
 
       ZM\Info("Login successful for user \"$username\"");
-      $password_type = password_type($user['Password']);
+      $password_type = password_type($user->Password());
 
       if ( $password_type == 'mysql' or $password_type == 'mysql+bcrypt' ) {
         ZM\Info('Migrating password, if possible for future logins');
@@ -378,10 +615,10 @@ if (ZM_OPT_USE_AUTH) {
       }
 
       if (ZM_AUTH_TYPE == 'builtin') {
-        $_SESSION['passwordHash'] = $user['Password'];
+        $_SESSION['passwordHash'] = $user->Password();
       }
 
-      $_SESSION['username'] = $user['Username'];
+      $_SESSION['username'] = $user->Username();
       if (ZM_AUTH_RELAY == 'plain') {
         // Need to save this in session, can't use the value in User because it is hashed
         $_SESSION['password'] = $_REQUEST['password'];
@@ -393,7 +630,7 @@ if (ZM_OPT_USE_AUTH) {
         $sql = 'SELECT * FROM Users WHERE Enabled=1 AND Username=?';
       }
       // local user, shouldn't affect the global user
-      $user = dbFetchOne($sql, NULL, array($_SERVER['REMOTE_USER']));
+      $user = new ZM\User(dbFetchOne($sql, NULL, array($_SERVER['REMOTE_USER'])));
     } else {
       $user = userFromSession();
     }
@@ -404,6 +641,7 @@ if (ZM_OPT_USE_AUTH) {
     }
   } # end if token based auth
 } else {
-  $user = $defaultUser;
+  global $defaultUser;
+  $user = new ZM\User($defaultUser);
 } # end if ZM_OPT_USE_AUTH
 ?>
